#!/usr/bin/perl
# -*- Perl -*-
# prnmz.pl - computation of PageRank(TM) for Namazu using GNU octave
# $Id: prnmz.pl,v 1.3 2001/02/28 09:48:25 baba Exp $
#
# Copyright (C) 2001 Hajime BABA  All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# This file must be encoded in EUC-JP encoding
#

package prnmz;

use strict;
use Cwd;
use File::Spec;
push(@INC, '/usr/share/namazu/pl');
require 'nmzidx.pl';
require 'util.pl';

my $octavepath = util::checkcmd('octave');
unless (defined $octavepath) {
    print STDERR "Octave is required to compute PageRank.\n";
    exit(1);
}

my $MEMORY_MAX = 32000000; # about 32MB for 2000 x 2000 array
my $solution = solution1(); # default: conservative
my $datafile = "NMZ.field.adjacency";
my $scriptfile = "NMZ.octave.m";
my $logfile = "NMZ.octave.log";
my $resultfile = "NMZ.octave.pagerank";

my $backup;
while (@ARGV && $ARGV[0] =~ s/^\-//){
    my $argv = shift;

    &usage, exit if $argv eq '-help';
    $backup = 0, next if $argv eq '-no-backup';
    $solution = solution1(), next if $argv eq '-1';
    $solution = solution2(), next if $argv eq '-2';

    while ($argv =~ s/^(.)//){
        $backup = 0 if $1 eq 'b';
    }
}

if (@ARGV){
    for my $argv (@ARGV){
        $argv =~ s/NMZ$// unless -d $argv;
        $argv = '.' if $argv eq '';
        &prnmz($argv, $backup) if -d $argv;
    }
}else{
    &prnmz('.', $backup);
}

exit(0);


sub prnmz {
    my ($dir, $backup) = @_;

    # select a solution.
    #select_solution($dir);

    # make NMZ.octave.m, which reads NMZ.field.adjacency
    make_script();
    
    # run Octave, and make NMZ.octave.pagerank
    system("$octavepath $scriptfile > $logfile");

    # read NMZ.octave.pagerank
    my @rank;
    open(F, "$resultfile") || die;
    while (<F>) {
	next if (/^#/);
	s/^\s+//g;
	push(@rank, $_);
    }
    close(F);

    # make NMZ.field.pagerank
    my $nmzo = new nmzidx($dir, 'w');
    my $fo = $nmzo->open_field('pagerank');
    foreach my $rank (@rank) {
	$fo->{'pagerank'}->putline($rank);
    }
    $nmzo->replace_db($backup);
}	


sub select_solution {
    my ($dir) = @_;

    # pretreatment: read NMZ.r, count $ndoc, then decide a solution.
    my $ndoc;
    my $nmzr = new nmzidx($dir, 'r');
    my $fr = $nmzr->open_flist;
    my %list_f;
    while (defined $fr->read(\%list_f)){
	$ndoc++;
    }
    $nmzr->close;
    if ($MEMORY_MAX < 8 * $ndoc * $ndoc) {
	$solution = solution1(); # slow, but conservative and economic.
    } else {
	$solution = solution2(); # fast, but needs more memory.
    }
}

sub make_script {
    my $script = "##
## NMZ.octave.m - GNU Octave script to compute PageRank(TM),
##                generated automatically by prnmz.pl.
##
## Copyright (C) 2001 Hajime BABA  All rights reserved.
##     This is free software with ABSOLUTELY NO WARRANTY.
##

##  This routine drives a number of powerful matrix manipulation
##  functions of Octave in order to optimize at full stretch.  It looks
##  so magical but the state-of-the-art routine is OK without
##  pretreatment of eliminating dangling pages.
##  Of course, the PageRanks of all of dangling pages are set to zero.
##
##  For the detailed algorithm of this script, see the reference:
##  Haveliwala, 'Efficient Computation of Pagerank', Stanford Technical
##  Report, 1999, http://dbpubs.stanford.edu:8090/pub/1999-31
## 
##  For the basic algorithm of PageRank, see the reference paper:
##  Page et al. (1998), 'The PageRank Citation Ranking: Bringing Order
##  to the Web', http://www-db.stanford.edu/~backrub/pageranksub.ps


## Set a wall-clock timer.
tic();

## Set a stochastic transition variable (aka. dumping factor).
c = 0.85;  # 0.85 is typical.  Must be set between 0 and 1.

## Make a sparse matrix using CRS (Compressed Row Storage) format.
## col_ind and row_ptr.  See the below URL in detail:
## http://www.netlib.org/linalg/old_html_templates/subsection2.8.3.1.html
fname = '$datafile'
[fid, msg] = fopen(fname, 'r');
if (fid == -1)
    fprintf(stderr, \"%s: %s\\n\", fname, msg);
    exit;
endif
col_ind = [];
row_ptr = [ 1 ];
r=0;
empty_list_elements_ok;
while (isstr(buf = fgets(fid)))
    r = r+1;
    tmp = sscanf(buf, '%d');
    tmp = tmp(find(tmp < r | tmp > r)); # squeeze loopback elements
    ## CRS format
    if (! isempty(tmp))
	if (isempty(col_ind))    
	    col_ind = [ tmp ];
	else
	    col_ind = [ col_ind ; tmp ];
	endif
    endif
    row_ptr(r+1) = row_ptr(r) + rows(tmp);
end
fclose(fid);
N=r;

## Define the other variables
HasLinks = diff(row_ptr) > 0; # bitmap column vector (sizeof Nx1)
for r=1:N
    IsLinked(r) = ! isempty(find(col_ind == r));
end
IsConnected = HasLinks & IsLinked; # bitmap column vector (sizeof Nx1)
N2 = length(find(IsConnected)); # number of pages which has (in|out)ward link
c2 = (1 - c) ./ N2;          # transition probability to other pages
Source = IsConnected ./ N2;  # initial vector of computation
eps = 1.0e-10;               # precision, overriding built-in value: 2.2e-16

## This is one of the solutions.
$solution
   
## Write the result to temporary file ($resultfile)
save -ascii $resultfile Dest;

## Check a wall-clock timer.
elapsed_time = toc()

## Check internal-use variables.
whos

## Local Variables:
## mode:octave-mode
## octave-block-offset:4
## End:
";

    open(F, "> $scriptfile") || die;
    print F "$script\n";
    close(F);

    return;
}


sub solution1 {
    my $solution = "
## Compute a PageRank vector.  Based on the iteration method after the
## manner of Section 3.1 of Haveliwala (1999), which is quite simple
## but works with lesser memory than that of original iteration method
## after the manner of Page et al. (1998).
## And the result vector, Dest, is a PageRank vector itself, indeed!
old_residual = 1.0;
residual = 1.0;
count = 0;
maxcount = 100;
while (residual > eps)
    count = count + 1;
    old_residual = residual;

    Dest = zeros(size(Source));
    for r = 1:N
	tmp = col_ind(row_ptr(r) : row_ptr(r+1)-1); # r-th row of datafile
	if (isempty(tmp)) continue; endif
	destvec = tmp(find(IsConnected(tmp))); # squeeze IsConnected=0 elements
	if (isempty(destvec)) continue; endif
	Dest(destvec) = Dest(destvec) + Source(r) ./ length(destvec);
    end
    Dest = c .* Dest + c2 .* IsConnected; # c2 = (1 - c) ./ N2
    residual = norm(Dest - Source, 1)     # L1 norm
    Source = Dest;

    ## stopping criterion
    if (count > maxcount)
	fprintf(stderr, \"Warning: maximum count.  Exit.\\n\");
	break;
    endif
    if ((residual == old_residual) | isinf(residual) | isnan(residual))
	fprintf(stderr, \"Warning: loop detected.  Exit.\\n\");
	break;
    endif
##    fprintf(stderr, \"%d, %g, %g\\n\", count, old_residual, residual);
end
count
norm(Dest,1)
";
    return $solution;
}

sub solution2 {
    my $solution = "
## Define stochastic matrix M, which size is NxN.  This is sparse.
M = zeros(N,N);
for r = 1:N
    tmp = col_ind(row_ptr(r) : row_ptr(r+1)-1); # r-th row of datafile
    if (isempty(tmp)) continue; endif
    destvec = tmp(find(IsConnected(tmp))); # squeeze IsConnected=0 elements
    if (isempty(destvec)) continue; endif
    M(r, destvec) = 1 ./ length(destvec);
end
M = c .* M' + c2 * IsConnected * IsConnected';

## Compute an eigenvector of A using the iteration method in Page et
## al. (1998), which is quite simple but works good in small N.  And
## the result is a PageRank vector itself, indeed!
residual = 1.0;
old_residual = 0.0;
count = 0;
maxcount = 100;
while (residual > eps)
    count = count + 1;
    old_residual = residual;

    Dest = M * Source;
    d = norm(Source, 1) - norm(Dest, 1);  # L1 norm
    Dest = Dest + d .* c2 .* IsConnected; # c2 = (1 - c) ./ N2
    residual = norm(Dest - Source, 1)     # L1 norm
    Source = Dest;

    ## stopping criterion
    if (count > maxcount)
	fprintf(stderr, \"Warning: maximum count.  Exit.\\n\");
	break;
    endif
    if ((residual == old_residual) | isinf(residual) | isnan(residual))
	fprintf(stderr, \"Warning: loop detected.  Exit.\\n\");
	break;
    endif
##    fprintf(stderr, \"%d, %g, %g\\n\", count, old_residual, residual);
end
count
norm(Dest,1)
";
    return $solution;
}

# EOF
